#include <stdlib.h>
#include <assert.h>
#include "basic.h"
#include "monitor.h"

monitor_t *alloc_monitor(compose_t f,uintmax_t r) {
  monitor_t *p = malloc_check(sizeof(*p)) ;
  p->lock = alloc_mutex() ;
  p->cond = alloc_cond() ;
  p->nrun = 0 ;
  p->r = r ;
  p->compose = f ;
  p->waiting = 0 ;
  return p ;
}

void free_monitor(monitor_t *p) {
  free_mutex(p->lock) ;
  free_cond(p->cond) ;
  free(p) ;
}

void enter_monitor(monitor_t *m) {
  lock_mutex(m->lock) ;
  m->nrun++ ;  
  unlock_mutex(m->lock) ;
}

void leave_monitor(monitor_t *m,uintmax_t y) {
  lock_mutex(m->lock) ;
  m->nrun-- ;
  m->r = m->compose(y,m->r) ;
  if (m->waiting && m->nrun == 0) signal_cond(m->cond) ;
  unlock_mutex(m->lock) ;
}

uintmax_t wait_monitor(monitor_t *m,int ntasks) {
  uintmax_t r ;
  lock_mutex(m->lock) ;
  m->waiting = 1 ;
  while (m->nrun > 0) wait_cond(m->cond,m->lock) ;
  m->waiting = 0 ;
  r = m->r ;
  unlock_mutex(m->lock) ;
  return r ;
}
