#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "basic.h"
#include "stack.h"

/* Alloc/free */
stack_t *alloc_stack(int sz) {
  stack_t *p = malloc_check(sizeof(*p));
  /* Stack array */
  p->sz = sz ;
  p->sp = 0 ;
  p->t = calloc_check(sz,sizeof(p->t[0]));
  /* Kill */
  p->killed = 0 ;
  p->seen = 0 ;
  /* Lock and conditions */
  p->lock = alloc_mutex();
  p->is_empty = alloc_cond();
  p->is_full = alloc_cond();
  p->wait = alloc_cond();
  return p ;
}

void free_stack(stack_t *p) {
  free(p->t);
  free_mutex(p->lock);
  free_cond(p->is_empty);
  free_cond(p->is_full);
  free_cond(p->wait);
  free(p);
}

static void stack_error(char *msg) {
  fprintf(stderr,"stack: %s\n",msg) ;
  exit(2);
}
/* push/pop */

void push(stack_t *p, void *q) {
  lock_mutex(p->lock) ;
  if (p->killed) stack_error("pushing on a killed stack");
  while (p->sp >= p->sz) wait_cond(p->is_full,p->lock) ;
  int was_empty = p->sp <= 0 ;
  p->t[p->sp] = q ; p->sp++ ;
  if (was_empty) broadcast_cond(p->is_empty) ;
  unlock_mutex(p->lock) ;
}


void *pop(stack_t *p) {
  void *r ;
  lock_mutex(p->lock) ;
  while (!p->killed && p->sp <= 0) wait_cond(p->is_empty,p->lock) ;
  if (p->sp > 0) {
    int was_full = p->sp >= p->sz ;
    p->sp-- ; r = p->t[p->sp] ;
    if (was_full) broadcast_cond(p->is_full) ;
  } else {
    assert (p->killed) ;
    p->seen++ ;
    signal_cond(p->wait) ;
    r = NULL ;
  }
  unlock_mutex(p->lock) ;
  return r ;
}

/* Kill */

void kill(stack_t *p,int nprocs) {
  lock_mutex(p->lock) ;
  if (p->killed) stack_error("killing the stack more then once");
  p->killed = 1 ;
  broadcast_cond(p->is_empty) ;
  while (p->seen < nprocs) wait_cond(p->wait,p->lock) ;
  unlock_mutex(p->lock) ;
}
